@@ -1,12 +1,12 @@
 /* Built-in and inline functions for gcj
-   Copyright (C) 2001, 2003, 2004, 2005
+   Copyright (C) 2001, 2003, 2004, 2005, 2006, 2007
    Free Software Foundation, Inc.
 
 This file is part of GCC.
 
 GCC is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
-the Free Software Foundation; either version 2, or (at your option)
+the Free Software Foundation; either version 3, or (at your option)
 any later version.
 
 GCC is distributed in the hope that it will be useful,
@@ -14,10 +14,9 @@ but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.
 
-You should have received a copy of the GNU General Public License
-along with GCC; see the file COPYING.  If not, write to
-the Free Software Foundation, 59 Temple Place - Suite 330,
-Boston, MA 02111-1307, USA.  
+.
+
+.  
 
 Java and all Java-based marks are trademarks or registered trademarks
 of Sun Microsystems, Inc. in the United States and other countries.
@@ -39,10 +38,9 @@ The Free Software Foundation is independent of Sun Microsystems, Inc.  */
 static tree max_builtin (tree, tree);
 static tree min_builtin (tree, tree);
 static tree abs_builtin (tree, tree);
+static tree convert_real (tree, tree);
 
 static tree java_build_function_call_expr (tree, tree);
-static void define_builtin (enum built_in_function, const char *,
-			    tree, const char *);
 
 
 
@@ -87,6 +85,10 @@ static GTY(()) struct builtin_record java_builtins[] =
   { { "java.lang.Math" }, { "sin" }, NULL, BUILT_IN_SIN },
   { { "java.lang.Math" }, { "sqrt" }, NULL, BUILT_IN_SQRT },
   { { "java.lang.Math" }, { "tan" }, NULL, BUILT_IN_TAN },
+  { { "java.lang.Float" }, { "intBitsToFloat" }, convert_real, 0 },
+  { { "java.lang.Double" }, { "longBitsToDouble" }, convert_real, 0 },
+  { { "java.lang.Float" }, { "floatToRawIntBits" }, convert_real, 0 },
+  { { "java.lang.Double" }, { "doubleToRawLongBits" }, convert_real, 0 },
   { { NULL }, { NULL }, NULL, END_BUILTINS }
 };
 
@@ -96,24 +98,30 @@ static GTY(()) struct builtin_record java_builtins[] =
 static tree
 max_builtin (tree method_return_type, tree method_arguments)
 {
-  return fold (build2 (MAX_EXPR, method_return_type,
-		       TREE_VALUE (method_arguments),
-		       TREE_VALUE (TREE_CHAIN (method_arguments))));
+  /* MAX_EXPR does not handle -0.0 in the Java style.  */
+  if (TREE_CODE (method_return_type) == REAL_TYPE)
+    return NULL_TREE;
+  return fold_build2 (MAX_EXPR, method_return_type,
+		      TREE_VALUE (method_arguments),
+		      TREE_VALUE (TREE_CHAIN (method_arguments)));
 }
 
 static tree
 min_builtin (tree method_return_type, tree method_arguments)
 {
-  return fold (build2 (MIN_EXPR, method_return_type,
-		       TREE_VALUE (method_arguments),
-		       TREE_VALUE (TREE_CHAIN (method_arguments))));
+  /* MIN_EXPR does not handle -0.0 in the Java style.  */
+  if (TREE_CODE (method_return_type) == REAL_TYPE)
+    return NULL_TREE;
+  return fold_build2 (MIN_EXPR, method_return_type,
+		      TREE_VALUE (method_arguments),
+		      TREE_VALUE (TREE_CHAIN (method_arguments)));
 }
 
 static tree
 abs_builtin (tree method_return_type, tree method_arguments)
 {
-  return fold (build1 (ABS_EXPR, method_return_type,
-		       TREE_VALUE (method_arguments)));
+  return fold_build1 (ABS_EXPR, method_return_type,
+		      TREE_VALUE (method_arguments));
 }
 
 /* Mostly copied from ../builtins.c.  */
@@ -123,20 +131,28 @@ java_build_function_call_expr (tree fn, tree arglist)
   tree call_expr;
 
   call_expr = build1 (ADDR_EXPR, build_pointer_type (TREE_TYPE (fn)), fn);
-  call_expr = build3 (CALL_EXPR, TREE_TYPE (TREE_TYPE (fn)),
+  return fold_build3 (CALL_EXPR, TREE_TYPE (TREE_TYPE (fn)),
 		      call_expr, arglist, NULL_TREE);
-  TREE_SIDE_EFFECTS (call_expr) = 1;
-  return fold (call_expr);
+}
+
+static tree
+convert_real (tree method_return_type, tree method_arguments)
+{
+  return build1 (VIEW_CONVERT_EXPR, method_return_type,
+		 TREE_VALUE (method_arguments));
 }
 
 
 
+#define BUILTIN_NOTHROW 1
+#define BUILTIN_CONST 2
 /* Define a single builtin.  */
 static void
 define_builtin (enum built_in_function val,
 		const char *name,
 		tree type,
-		const char *libname)
+		const char *libname,
+		int flags)
 {
   tree decl;
 
@@ -144,10 +160,13 @@ define_builtin (enum built_in_function val,
   DECL_EXTERNAL (decl) = 1;
   TREE_PUBLIC (decl) = 1;
   SET_DECL_ASSEMBLER_NAME (decl, get_identifier (libname));
-  make_decl_rtl (decl);
   pushdecl (decl);
   DECL_BUILT_IN_CLASS (decl) = BUILT_IN_NORMAL;
   DECL_FUNCTION_CODE (decl) = val;
+  if (flags & BUILTIN_NOTHROW)
+    TREE_NOTHROW (decl) = 1;
+  if (flags & BUILTIN_CONST)
+    TREE_READONLY (decl) = 1;
 
   implicit_built_in_decls[val] = decl;
   built_in_decls[val] = decl;
@@ -187,44 +206,62 @@ initialize_builtins (void)
   double_ftype_double_double = build_function_type (double_type_node, t);
 
   define_builtin (BUILT_IN_FMOD, "__builtin_fmod",
-		  double_ftype_double_double, "fmod");
+		  double_ftype_double_double, "fmod", BUILTIN_CONST);
   define_builtin (BUILT_IN_FMODF, "__builtin_fmodf",
-		  float_ftype_float_float, "fmodf");
+		  float_ftype_float_float, "fmodf", BUILTIN_CONST);
 
   define_builtin (BUILT_IN_ACOS, "__builtin_acos",
-		  double_ftype_double, "_ZN4java4lang4Math4acosEd");
+		  double_ftype_double, "_ZN4java4lang4Math4acosEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_ASIN, "__builtin_asin",
-		  double_ftype_double, "_ZN4java4lang4Math4asinEd");
+		  double_ftype_double, "_ZN4java4lang4Math4asinEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_ATAN, "__builtin_atan",
-		  double_ftype_double, "_ZN4java4lang4Math4atanEd");
+		  double_ftype_double, "_ZN4java4lang4Math4atanEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_ATAN2, "__builtin_atan2",
-		  double_ftype_double_double, "_ZN4java4lang4Math5atan2Edd");
+		  double_ftype_double_double, "_ZN4java4lang4Math5atan2EJddd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_CEIL, "__builtin_ceil",
-		  double_ftype_double, "_ZN4java4lang4Math4ceilEd");
+		  double_ftype_double, "_ZN4java4lang4Math4ceilEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_COS, "__builtin_cos",
-		  double_ftype_double, "_ZN4java4lang4Math3cosEd");
+		  double_ftype_double, "_ZN4java4lang4Math3cosEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_EXP, "__builtin_exp",
-		  double_ftype_double, "_ZN4java4lang4Math3expEd");
+		  double_ftype_double, "_ZN4java4lang4Math3expEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_FLOOR, "__builtin_floor",
-		  double_ftype_double, "_ZN4java4lang4Math5floorEd");
+		  double_ftype_double, "_ZN4java4lang4Math5floorEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_LOG, "__builtin_log",
-		  double_ftype_double, "_ZN4java4lang4Math3logEd");
+		  double_ftype_double, "_ZN4java4lang4Math3logEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_POW, "__builtin_pow",
-		  double_ftype_double_double, "_ZN4java4lang4Math3powEdd");
+		  double_ftype_double_double, "_ZN4java4lang4Math3powEJddd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_SIN, "__builtin_sin",
-		  double_ftype_double, "_ZN4java4lang4Math3sinEd");
+		  double_ftype_double, "_ZN4java4lang4Math3sinEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_SQRT, "__builtin_sqrt",
-		  double_ftype_double, "_ZN4java4lang4Math4sqrtEd");
+		  double_ftype_double, "_ZN4java4lang4Math4sqrtEJdd",
+		  BUILTIN_CONST);
   define_builtin (BUILT_IN_TAN, "__builtin_tan",
-		  double_ftype_double, "_ZN4java4lang4Math3tanEd");
+		  double_ftype_double, "_ZN4java4lang4Math3tanEJdd",
+		  BUILTIN_CONST);
   
   t = tree_cons (NULL_TREE, boolean_type_node, end_params_node);
   t = tree_cons (NULL_TREE, boolean_type_node, t);
   boolean_ftype_boolean_boolean = build_function_type (boolean_type_node, t);
   define_builtin (BUILT_IN_EXPECT, "__builtin_expect", 
 		  boolean_ftype_boolean_boolean,
-		  "__builtin_expect");
+		  "__builtin_expect",
+		  BUILTIN_CONST | BUILTIN_NOTHROW);
 		  
+  define_builtin (BUILT_IN_SYNCHRONIZE, "__sync_synchronize",
+		  build_function_type (void_type_node, void_list_node),
+		  "__sync_synchronize", BUILTIN_NOTHROW);
+
   build_common_builtin_nodes ();
 }
 
@@ -249,11 +286,20 @@ check_for_builtin (tree method, tree call)
 	      tree fn;
 
 	      if (java_builtins[i].creator != NULL)
-		return (*java_builtins[i].creator) (method_return_type,
-						    method_arguments);
+		{
+		  tree result
+		    = (*java_builtins[i].creator) (method_return_type,
+						   method_arguments);
+		  return result == NULL_TREE ? call : result;
+		}
+
+	      /* Builtin functions emit a direct call which is incompatible
+	         with the BC-ABI.  */
+	      if (flag_indirect_dispatch)
+	        return call;
 	      fn = built_in_decls[java_builtins[i].builtin_code];
 	      if (fn == NULL_TREE)
-		return NULL_TREE;
+		return call;
 	      return java_build_function_call_expr (fn, method_arguments);
 	    }
 	}
